// ######################################################################
//
//  Target Tracker File v1.0
//
//  Released:  08/12/2015
//
//  Author: TechKnowTone
//
// ######################################################################
/*
    This software is furnished "as is", without technical support, and
    with no warranty, expressed or implied, as to its usefulness for any
    purpose.
    
    Elements of this software were taken from the public domain. It uses
    libraries for tone generation, servo motor control and multi-coloured
    LEDs.
    
    This program drives the twin servo robot, with ultrasonic radar,
    and audible noises. it also confirms that the servo are working and
    allows you to set a specific angle from the keyboard for each. Now
    a single pixel LED can be controlled. Now includes ultrasonic
    range finder. This version is for the complete vehicle and it now
    responds to the remote 'keyboard' Nxx. commands.

    Commands:
    AU0.    - set audio OFF
    AU1.    - set audio ON
    DPnn.   - set the degrees/pulse to nn
    FLnn.   - set left Fire LED to brightness nn
    FRnn.   - set right Fire LED to brightness nn
    GA0.    - set gun inactive
    GA1.    - set gun active
    GC.     - move gun to centre
    GLnn.   - sets gun lower angle limit to nn
    GMnnn.  - move gun to angle nnn
    GS.     - toggle gun sweep mode
    GUnn.   - sets gun upper angle limit to nn
    GZ.     - switch OFF gun servo drive
    M1.     - mode 1
    M2.     - mode 2.
    M3.     - mode 3.
    Nnn.    - play a continuous note nn until N0. is received
    PBbb.   - sets LED Blue value to nn
    PGbb.   - sets LED Green value to nn
    PRbb.   - sets LED Red value to nn
    PZ.     - turns OFF Pixel LED
    RC.     - move radar to centre
    RLnn.   - sets radar lower angle limit to nn
    RMnnn.  - move radar to angle nnn
    RS0.    - turn radar sweep OFF
    RS1.    - turn radar sweep ON
    RUnn.   - sets gun upper angle limit to nn
    RZ.     - switch OFF radar servo drive
    TFnn.   - set tone frequency to nn
    TPnn.   - tune play melody(nn)
    TZ.     - turn tone OFF
    U0.     - turn ultrasonic radar OFF
    U1.     - turn ultrasonic radar ON
    Wnnn.   - wait between do commands for nn x 10ms
    XX.     - Processing application is exiting, so go safe
    Znn.    - used in development to set a given variable

*/
// Declare servo objects
#include <Adafruit_NeoPixel.h>
#include "pitches.h"
#include <Servo.h>

Servo gunServo;  // create servo object for gun servo
Servo radServo;  // create servo object for radar servo

// define labels and constants
#define gunCtrCntDef 5 // gun to centre counter default
#define gunLEDLft 9 // LH gun LED
#define gunLEDRht 10 // RH gun LED
#define gunServoPin 6 // gun servo pin
#define ledPin 12 // RGB LED serial pin
#define modeSwitch 11 //mode switch i/p assignment
#define radServoPin 7 // radar servo pin
#define servoGMaxMax 135  // hard limit for servoGMax
#define servoGMinMin 45  // hard limit for servoGMin
#define servoRMaxMax 150  // hard limit for servoRMax
#define servoRMinMin 30  // hard limit for servoRMin
#define tonePinHi 2 // high output audio pin assignment
#define tonePinLo 3 // low output audio pin assignment
#define trigPin 5   // ultrasonic trigger pin

// Parameter 1 = number of pixels in strip
// Parameter 2 = pin number (most are valid)
// Parameter 3 = pixel type flags, add together as needed:
//   NEO_RGB     Pixels are wired for RGB bitstream
//   NEO_GRB     Pixels are wired for GRB bitstream
//   NEO_KHZ400  400 KHz bitstream (e.g. FLORA pixels)
//   NEO_KHZ800  800 KHz bitstream (e.g. High Density LED strip)
Adafruit_NeoPixel strip = Adafruit_NeoPixel(3, ledPin, NEO_GRB + NEO_KHZ800);

// define arrays
#define arraySize 80 //max no. notes in a tune
#define fireSize 40 // max no. fires in a gun sequence
unsigned int fireArray[fireSize]; // fire states and durations gun fire
unsigned int melody[arraySize]; // notes and durations in melody

// Declare and initialise global variables
int AlertMax; // gun range alert max distance = I can see you!
int AlertMin; // gun range alert min distance = Fire!
int AudioEn; // = 1 for audio enable, 0 = OFF
uint8_t colBlu[] = {0, 0, 0}; //colour blue
uint8_t colGrn[] = {0, 0, 0}; //colour green
uint8_t colRed[] = {0, 0, 0}; //colour red
char cmdMode; // command mode G-gun, R-radar
char cmdType; // command mode type
int cmdVal; // value associated with a cmdType
long distance; // distance measured by radar
int doWait; // pause in 10ms increments during do commands
long duration;  // measured time delay from echo response
long echoMax;  // max pulse timeout 5 ms
int echoPin;   // ultrasonic echo pin
long fireDuration; // duration of a fire shot being played in 10ms
int firePnt; // set >= 0 to fire gun sequence
int gunActEn; // gun active/inactive flag
int gunCtrCnt; // use to slowly centralise gun
unsigned long interval; // main loop interval in microseconds
char keyChar; // any keyboard character
int keyVal; // any keyboard value
long ledCnt; // RGB LED sequence counter
long ledCntSet; // reset value for ledCnt
int ledMode; // RGB LED mode state
int loadFire; // pointer used to load arrays
int loadPnt; // pointer used to load arrays
unsigned long nextMicros; // main loop interval in microseconds
int noteDuration; // duration of a note being played in 10ms
int noteDurationVal; // reference duration of a note being played
int notePause; // inter-note pause counter
int numFires;  // number of fires. Must be less than fireSize
int numNotes;  // number of notes in a melody. Must be less than arraySize
int onceFlag = 1; // runs events once in POST
int playPnt; // set >= 0 to play melody sequence
int radarCnt; // counter load value for radar trig time-out in 10ms
int radarFlag; // = 1 if sweeping up, -1 if sweeping down
int radarPnt; // current gun sweep value
int radarTrig; // radar trigger/counter
int servoGCtr; // gun centre servo angle limit
int servoGEn; // gun servo enable, 0 = OFF, 1 = ON
int servoGMax; // gun max servo angle limit
int servoGMin; // gun min servo angle limit
int sevoGOff; // offset required to achieve centre alignment
int servoGVal; // latest value sent to the gun servo
int servoRCtr; // radar centre servo angle limit
int servoREn; // radar servo enable, 0 = OFF, 1 = ON
int servoRMax; // radar max servo angle limit
int servoRMin; // radar min servo angle limit
int servoROff; // offset require to achieve centre alignment
int servoRVal; // latest value sent to the radar servo
int swDel; // delay between siwtch reads
int swCnt; // number of switch presses
int sweepFlag; // = 1 if sweeping up, -1 if sweeping down
int sweepPnt; // current gun sweep value
int swRead; // latest input value
int swState; //previous switch state
int swTimeout; // inactivity timeout timer
int swVal; // temporary switch value or inactivity timer
int targetA0; // first detected target angle
int targetA1; // lastst detected target angle
int targetAM; // derived target angle
int toneFreq; // frequency of note being played
int toneFreq0; // note frequency for single note
int toneFreq1; // target note frequency for swept note
int tonePin; // default tone pin
int toneSwpF; // = 1 if swept tone is being played
int victoryCnt; // >1 count down then play tune when = 1


void setup() {
  // put your setup code here, to run once:
  setDefaults();
  pinMode(echoPin, INPUT);
  pinMode(gunLEDLft, OUTPUT); digitalWrite(gunLEDLft, LOW);
  pinMode(gunLEDRht, OUTPUT); digitalWrite(gunLEDRht, LOW);
  pinMode(gunServoPin, OUTPUT);
  pinMode(modeSwitch, INPUT_PULLUP);
  pinMode(radServoPin, OUTPUT);
  pinMode(trigPin, OUTPUT);
  gunServo.attach(gunServoPin);  // attaches gun servo to pin 6
  radServo.attach(radServoPin);  // attaches radar servo to pin 7
  strip.begin();
  strip.show(); // Initialize all pixels to 'off'
  Serial.begin(19200);
  runPOST();
}

void loop() {
  // Main loop runs on a 10 ms time interval
  if ((playPnt != -1) && (AudioEn > 0)) {playATune();}
  if (firePnt != -1) {fireGun();}
  if (sweepFlag != 0) {sweepGun();}
  if (radarFlag != 0) {sweepRadar();}
  if (radarTrig > 0) {
    if (radarTrig == 1) {trigRadar();}
    else if (firePnt == -1){moveGunCtr();}
    radarTrig--;
    if (radarTrig < 1) {radarTrig = radarCnt;}
  }
  keyVal = 0; while (keyVal != -1) {readKey();} // empty Rx buffer
  swDel--; if (swDel <1) {readModeSwitch();} // test mode switch
  if (victoryCnt > 0) {victoryCnt--; if (victoryCnt == 1) {loadTuneVictory();}}
  if (ledCnt > 0) {ledSeq();}
  // main loop time adjustment.
  while (micros() < nextMicros) {
    // wait for remainder of 10ms period to sync overall period
  } nextMicros = micros() + interval;
}

// ----------------------------------------------------------------------

void doCmd() {
  // a '.' has been received so execute command if valid
  Serial.print(cmdMode); Serial.print(cmdType); Serial.print(cmdVal); Serial.println('.');
  if (doWait < 1) {
    switch (cmdMode) {
      case ' ': break;
      case 'A':
        switch (cmdType) {
          case ' ': break;
          case 'U': setAudioFlag(cmdVal); break;
        } break;
      case 'D':
        switch (cmdType) {
          case ' ': break;
          case 'P': setDP(cmdVal); break;
        } break;
      case 'F':
        switch (cmdType) {
          case ' ': break;
          case 'L': setFireLft(cmdVal); break;
          case 'R': setFireRht(cmdVal); break;
        } break;
      case 'G':
        switch (cmdType) {
          case ' ': break;
          case 'A': setGunActive(cmdVal); break;
          case 'C': moveGunServo(servoGCtr); break;
          case 'L': setGunServoMin(cmdVal); break;
          case 'M': moveGunServo(cmdVal); break;
          case 'S': toggleGunSweep(); break;
          case 'U': setGunServoMax(cmdVal); break;
          case 'Z': setGunServoOFF(); break;
        } break;
      case 'M':
        switch (cmdVal) {
          case 1: swCnt = 1; modeKeyAction(); break;
          case 2: swCnt = 2; modeKeyAction(); break;
          case 3: swCnt = 3; modeKeyAction(); break;
        } break;
      case 'N':
        playNote(cmdVal); break;
      case 'P':
        switch (cmdType) {
          case ' ': break;
          case 'B': setPixelBlu(cmdVal); break;
          case 'G': setPixelGrn(cmdVal); break;
          case 'R': setPixelRed(cmdVal); break;
          case 'Z': setPixelOFF(); break;
        } break;
      case 'R':
        switch (cmdType) {
          case ' ': break;
          case 'C': moveRadServo(servoRCtr); break;
          case 'L': setRadServoMin(cmdVal); break;
          case 'M': moveRadServo(cmdVal); break;
          case 'S': toggleRadSweep(); break;
          case 'U': setRadServoMax(cmdVal); break;
          case 'Z': setRadServoOFF(); break;
        } break;
      case 'T':
        switch (cmdType) {
          case ' ': break;
          case 'F': setToneFreq(cmdVal); break;
          case 'P':
            switch(cmdVal) {
              case 0: loadTune0(); break;
            } break;
          case 'Z': noTone(tonePin); break;
        } break;
      case 'U':
        switch (cmdType) {
          case ' ': toggleRadarTrig(); break;
        } break;
      case 'W':
        doWait = cmdVal; break;
      case 'X':
        switch (cmdType) {
          case ' ': break;
          case 'X': goSafe(); break;
        } break;
      case 'Z':
        // used for deveopment and tuning
        ledCntSet = cmdVal; break;
    }
    // now reset the variables
    cmdMode = ' '; cmdType = ' '; cmdVal = 0;
  } else {doWait--;}
}

// ----------------------------------------------------------------------

void extendCmdVal(int zVal) {
  // adds a new digit to the right-hand end of cmdVal
  cmdVal = (cmdVal * 10) + zVal;  
}

// ----------------------------------------------------------------------

void fireGun() {
  //   play a fire sequence
  if (fireDuration > 0) {
    fireDuration--; // already flashing so count down
  }
  else {
    if (firePnt > -1) {
//      fireState = fireArray[firePnt]; firePnt++;
//      int zS = LOW: if (fireState > 0) {zS = HIGH;}
      int zS = fireArray[firePnt]; firePnt++;
      digitalWrite(gunLEDLft, zS);
      digitalWrite(gunLEDRht, zS);
      fireDuration = fireArray[firePnt]; firePnt++;
      if (firePnt >= numFires) {firePnt = -2;} // reached finel fire
    } else if (firePnt == -2) {
    // last element to turn LEDs OFF
    digitalWrite(gunLEDLft, LOW); digitalWrite(gunLEDRht, LOW);
    firePnt = -1; 
    }
  }
}

// ----------------------------------------------------------------------

void goSafe() {
  // received XX. from 'Processing' so go safe
  moveRadServo(servoRCtr);
  moveGunServo(servoGCtr);
  playPnt = -1; noTone(tonePin);
  digitalWrite(gunLEDLft, LOW); digitalWrite(gunLEDRht, LOW);
  sweepFlag = 0;
  radarFlag = 0;
  radarTrig = 0;
}

// ----------------------------------------------------------------------

void ledSeq() {
  // sequence the RGB LEDs
  ledCnt--;
  if (ledCnt < 1) {
    ledCnt = ledCntSet;
    switch (ledMode) {
      case 0: ledMode = 1; setPixiCol(0,0,40,0); setPixiCol(1,0,0,0); setPixiCol(2,0,15,0); 
        break;
      case 1: ledMode = 2; setPixiCol(0,0,15,0); setPixiCol(1,0,40,0); setPixiCol(2,0,0,0); 
        break;
      case 2: ledMode = 0; setPixiCol(0,0,0,0); setPixiCol(1,0,15,0); setPixiCol(2,0,40,0); 
        break;
      case 3: ledMode = 4; setPixiCol(0,1,0,0); setPixiCol(1,0,0,0); setPixiCol(2,1,0,0); 
        break;
      case 4: ledMode = 5; setPixiCol(0,5,0,0); setPixiCol(1,5,0,0); setPixiCol(2,5,0,0); 
        break;
      case 5: ledMode = 3; setPixiCol(0,0,0,0); setPixiCol(1,15,0,0); setPixiCol(2,0,0,0); 
        break;
      case 90: ledMode = 91;
        setPixiCol(0,random(20),random(20),random(20));
        break;
      case 91: ledMode = 92;
        setPixiCol(1,random(20),random(20),random(20));
        break;
      case 92: ledMode = 90;
        setPixiCol(2,random(20),random(20),random(20));
        break;
    }
  }
}

// ----------------------------------------------------------------------

void loadArrayFires(int zS, int zD) {
  // load values into fire array at position loadPnt
  fireArray[loadFire] = zS; loadFire++;
  fireArray[loadFire] = zD; loadFire++;
}

// ----------------------------------------------------------------------

void loadArrayFreq(int zF, int zD) {
  // load values into melody array at position loadPnt
  // if zD is negative a 20ms pause will be added to the end of the note
  melody[loadPnt] = zF; loadPnt++;
  melody[loadPnt] = zD; loadPnt++;
}

// ----------------------------------------------------------------------

void loadArraySweep(int zF1, int zF2, int zD) {
  // load values into melody array at position loadPnt
  // if zD is negative a 1/3 pause will be added to the end of the note
  melody[loadPnt] = -zF1; loadPnt++; // negative to indicate sweep!
  melody[loadPnt] = zF2; loadPnt++;
  melody[loadPnt] = zD; loadPnt++;
}

// ----------------------------------------------------------------------

void loadTune0() {
  // load tune notes and durations into the melody array
  // this tune: dah-dah
  noTone(tonePin); // stop current tune
  tonePin = tonePinLo; // set volume low
  loadPnt = 0;
  loadArrayFreq(NOTE_C5, -25);
  loadArrayFreq(NOTE_C4, -25);
  loadArrayFreq(NOTE_C5, -25);
  loadArrayFreq(0, 100);
  loadArrayFreq(NOTE_C5, 25);
  loadArrayFreq(NOTE_C4, 25);
  loadArrayFreq(NOTE_C5, 25);
  loadArrayFreq(0, 100);
  loadArraySweep(NOTE_C4, NOTE_C6, 10);
  loadArraySweep(NOTE_C4, NOTE_C6, 10);
  loadArraySweep(NOTE_C4, NOTE_C6, 10);
  loadArraySweep(NOTE_C6, NOTE_C4, 10);
  loadArraySweep(NOTE_C6, NOTE_C4, 10);
  loadArraySweep(NOTE_C6, NOTE_C4, 10);

  numNotes = loadPnt; // length of melody addressed, numNotes - 1
  playPnt = 0; // point to first note to start playing melody
}

// ----------------------------------------------------------------------

void loadTuneFire() {
  // load tune notes and durations into the melody array
  // this tune: ping!-ping!
  noTone(tonePin); // stop current tune
  tonePin = tonePinHi; // set volume high
  loadPnt = 0;
  loadArraySweep(NOTE_C4, NOTE_C6, 10);
  loadArraySweep(NOTE_C4, NOTE_C6, 10);
  loadArraySweep(NOTE_C6, NOTE_C4, 20);
  numNotes = loadPnt; // length of melody addressed, numNotes - 1
  playPnt = 0; // point to first note to start playing melody
  loadFire = 0;
  loadArrayFires(HIGH, 5);
  loadArrayFires(LOW, 5);
  loadArrayFires(HIGH, 5);
  loadArrayFires(LOW, 5);
  loadArrayFires(HIGH, 20);
  numFires = loadFire; // length of melody addressed, numNotes - 1
  firePnt = 0; // point to first elemnt to start playing fire sequence
}

// ----------------------------------------------------------------------

void loadTuneVictory() {
  // load tune notes and durations into the melody array
  // this tune: dah-da-da-da-dah-dah-dah!
  noTone(tonePin); // stop current tune
  tonePin = tonePinHi; // set volume low
  loadPnt = 0;
  loadArrayFreq(NOTE_C5, -25);
  loadArrayFreq(NOTE_C5, -10);
  loadArrayFreq(NOTE_C5, -10);
  loadArrayFreq(NOTE_C5, -10);
  loadArrayFreq(NOTE_D5, -20);
  loadArrayFreq(NOTE_C5, -20);
  loadArrayFreq(NOTE_F5, -30);

  numNotes = loadPnt; // length of melody addressed, numNotes - 1
  playPnt = 0; // point to first note to start playing melody
}

// ----------------------------------------------------------------------

void loadTuneWarning() {
  // load tune notes and durations into the melody array
  // this tune: ping!-ping!
  noTone(tonePin); // stop current tune
  tonePin = tonePinLo; // set volume low
  loadPnt = 0;
  loadArrayFreq(NOTE_C5, 25);
  loadArrayFreq(NOTE_C4, 25);

  numNotes = loadPnt; // length of melody addressed, numNotes - 1
  playPnt = 0; // point to first note to start playing melody
}

// ----------------------------------------------------------------------

void modeKeyAction() {
  // change mode depending on swCnt value
  switch (swCnt) {
    case 1: resetAll(); break;
    case 2: setMode(2); break;
    case 3: setMode(3); break;
  } swCnt = 0; // clear count for next sequence
}

// ----------------------------------------------------------------------

void moveGunCtr() {
  // called to slowly move gun to centre position  
  gunCtrCnt--;
  if (gunCtrCnt < 1) {
    // determine direction to move?
    if (servoGVal > servoGCtr) {servoGVal--;}
    if (servoGVal < servoGCtr) {servoGVal++;}
    moveGunServo(servoGVal);
    gunCtrCnt = gunCtrCntDef; // reset counter
  }
}

// ----------------------------------------------------------------------

void moveGunServo(int zcmdVal) {
  // move the gun servo to a specific angle
  if (zcmdVal < servoGMin) {zcmdVal = servoGMin;}
  if (zcmdVal > servoGMax) {zcmdVal = servoGMax;}
  servoGVal = zcmdVal;
  gunServo.write(180 - servoGVal + sevoGOff);
  sweepFlag = 0; // stop sweep if active
}

// ----------------------------------------------------------------------

void moveRadServo(int zcmdVal) {
  // mode the radar servo to a specific angle
  if (zcmdVal < servoRMinMin) {zcmdVal = servoRMinMin;}
  if (zcmdVal > servoRMaxMax) {zcmdVal = servoRMaxMax;}
  servoRVal = zcmdVal;
  radServo.write(180 - servoRVal + servoROff);
  radarFlag = 0; // stop sweep if active
  Serial.print(">-1,");Serial.print(servoRVal);
  Serial.println(".");
}

// ----------------------------------------------------------------------

void playATune() {
  // playing a note sequence
  if (noteDuration > 0) {
    noteDuration--; // already playing a note so count down duration
    if (toneSwpF > 0) {
      int zFreq = toneFreq0 + (((toneFreq1 - toneFreq0) * (noteDurationVal - noteDuration))/noteDurationVal);
      if (zFreq != toneFreq) {setToneFreq(zFreq);}
    }
  }
  else if (notePause > 0) {
    notePause--; noTone(tonePin);
  }
  else {
    if (playPnt > -1) {
      toneFreq0 = melody[playPnt]; playPnt++; toneSwpF = 0;
      if (toneFreq0 < 0) {
        // negative freq indicates swept note
        toneFreq0 = -toneFreq0; toneSwpF = 1;
        toneFreq1 = melody[playPnt]; playPnt++;
      }
      setToneFreq(toneFreq0); 
      noteDuration = melody[playPnt]; playPnt++;
      if (noteDuration < 0) {
         // negative number so introduce an inter-note pause
        noteDuration = -noteDuration; notePause = noteDuration/3;
      } noteDurationVal = noteDuration;
      if (playPnt >= numNotes) {playPnt = -2;} // reached final note
    } else if (playPnt == -2) {
      noTone(tonePin); playPnt = -1; // last note ended so stop tune
    }
  }
}

// ----------------------------------------------------------------------

void playNote(int zFreq) {
  // play a note continuously
  if (playPnt > -1) {noTone(tonePin); playPnt = -1;}
  if (zFreq > 0) {
    // play the note continuously
    setToneFreq(zFreq);
  } else {noTone(tonePin);} // received 0 so no tone
}

// ----------------------------------------------------------------------

void readKey() {
  // reads a key from the keyboard annd reacts accordingly
  if (doWait < 1) {
    keyVal = Serial.read();
    if (keyVal != -1) {
      keyChar = char(keyVal);
//      Serial.print("keyVal=");
//      Serial.print(keyChar);
//      Serial.print("   ASCII=");
//      Serial.println(keyVal);
      int zNF = 0;
      switch (keyChar) {
        case '0': extendCmdVal(0); zNF = 1;break;
        case '1': extendCmdVal(1); zNF = 1;break;
        case '2': extendCmdVal(2); zNF = 1;break;
        case '3': extendCmdVal(3); zNF = 1;break;
        case '4': extendCmdVal(4); zNF = 1;break;
        case '5': extendCmdVal(5); zNF = 1;break;
        case '6': extendCmdVal(6); zNF = 1;break;
        case '7': extendCmdVal(7); zNF = 1;break;
        case '8': extendCmdVal(8); zNF = 1;break;
        case '9': extendCmdVal(9); zNF = 1;break;
      }
      if (zNF == 0) {
        if (cmdMode == ' ') {
          // test for new Command Mode char?
          switch (keyChar) {
            case 'a': cmdMode = 'A'; break;
            case 'A': cmdMode = 'A'; break;
            case 'd': cmdMode = 'D'; break;
            case 'D': cmdMode = 'D'; break;
            case 'f': cmdMode = 'F'; break;
            case 'F': cmdMode = 'F'; break;
            case 'g': cmdMode = 'G'; break;
            case 'G': cmdMode = 'G'; break;
            case 'm': cmdMode = 'M'; break;
            case 'M': cmdMode = 'M'; break;
            case 'n': cmdMode = 'N'; break;
            case 'N': cmdMode = 'N'; break;
            case 'p': cmdMode = 'P'; break;
            case 'P': cmdMode = 'P'; break;
            case 'r': cmdMode = 'R'; break;
            case 'R': cmdMode = 'R'; break;
            case 't': cmdMode = 'T'; break;
            case 'T': cmdMode = 'T'; break;
            case 'u': cmdMode = 'U'; break;
            case 'U': cmdMode = 'U'; break;
            case 'w': cmdMode = 'W'; break;
            case 'W': cmdMode = 'W'; break;
            case 'x': cmdMode = 'X'; break;
            case 'X': cmdMode = 'X'; break;
            case 'z': cmdMode = 'Z'; break;
            case 'Z': cmdMode = 'Z'; break;
          } cmdType = ' '; cmdVal = 0;
        } else {
          // test for Command Type char?
          switch (keyChar) {
            case 'a': cmdType = 'A'; break;
            case 'A': cmdType = 'A'; break;
            case 'b': cmdType = 'B'; break;
            case 'B': cmdType = 'B'; break;
            case 'c': cmdType = 'C'; break;
            case 'C': cmdType = 'C'; break;
            case 'f': cmdType = 'F'; break;
            case 'F': cmdType = 'F'; break;
            case 'g': cmdType = 'G'; break;
            case 'G': cmdType = 'G'; break;
            case 'l': cmdType = 'L'; break;
            case 'L': cmdType = 'L'; break;
            case 'm': cmdType = 'M'; break;
            case 'M': cmdType = 'M'; break;
            case 'p': cmdType = 'P'; break;
            case 'P': cmdType = 'P'; break;
            case 'r': cmdType = 'R'; break;
            case 'R': cmdType = 'R'; break;
            case 's': cmdType = 'S'; break;
            case 'S': cmdType = 'S'; break;
            case 'u': cmdType = 'U'; break;
            case 'U': cmdType = 'U'; break;
            case 'x': cmdType = 'X'; break;
            case 'X': cmdType = 'X'; break;
            case 'z': cmdType = 'Z'; break;
            case 'Z': cmdType = 'Z'; break;
            case '.': doCmd(); break;
          }
        }
      } else if (keyChar == '.') {doCmd();}
    }
  }
}

// --------------------------------------------------------------------

void readModeSwitch() {
  // reads a single switch and counts the number of presses
  swDel = 2; // come here every 20ms
  swRead = digitalRead(modeSwitch);
  if (swRead != swState) {
    // a transition has occured!
    if (swRead == LOW) {swCnt++; swVal = 0;}
  }
  swState = digitalRead(modeSwitch);
  swVal++; // increment timeout delay
  if (swVal == 50) {
    // no more presses after 1.0 seconds
//    Serial.print("swCnt: ");
//    Serial.println(swCnt);
    modeKeyAction();
  }
}

// ----------------------------------------------------------------------

void resetAll () {
  // called from modeSwitch clicked once
  noTone(tonePin);
  setDefaults();
  goSafe();
  setPixelOFF();
  runPOST();
}

// ----------------------------------------------------------------------

void runPOST() {
  // perform power-up tests here
  goSafe();
  for (int zL = 1; zL < 20; zL++) {Serial.println("");} // CLS
  if (onceFlag > 0) {
    // only do these things on power-up or USB reset
    setPixiCol(0,100,0,0); setPixiCol(1,50,50,0); setPixiCol(2,0,100,0);
    loadPnt = 0;
    tonePin = tonePinLo;
    loadArraySweep(NOTE_C4, NOTE_C6, 10);
    loadArraySweep(NOTE_C6, NOTE_C4, 10);
    numNotes = loadPnt; playPnt = 0;
    while (playPnt != -1) {playATune(); delay(10);}
    loadPnt = 0;
    tonePin = tonePinHi;
    loadArraySweep(NOTE_C4, NOTE_C6, 10);
    loadArraySweep(NOTE_C6, NOTE_C4, 10);
    numNotes = loadPnt; playPnt = 0;
    digitalWrite(gunLEDLft, HIGH); digitalWrite(gunLEDRht, HIGH);
    while (playPnt != -1) {playATune(); delay(10);}
    digitalWrite(gunLEDLft, LOW); digitalWrite(gunLEDRht, LOW);
  } onceFlag = 0;
//  setPixiCol(0,10,10,0); setPixiCol(1,10,10,0); setPixiCol(2,10,10,0);
  Serial.println("#POST Complete.");

  // set main loop timer before exiting
  tonePin = tonePinLo;
  setMode(1);
  nextMicros = micros() + interval;
}

// ----------------------------------------------------------------------

void setAudioFlag(int zcmdVal) {
  //toggle audio flag ON/OFF  
  if (zcmdVal > 0) {
    AudioEn = 1;
  } else {
    AudioEn = 0; noTone(tonePin); playPnt = -1;
  }
}

// ----------------------------------------------------------------------

void setDefaults() {
  // set default variable values
  AlertMax = 3000; // gun range alert max distance = I can see you!
  AlertMin = 2000; // gun range alert min distance = Fire!
  AudioEn = 1; // = 1 for audio enable, 0 = OFF
  colBlu[0] = 0; colBlu[1] = 0; colBlu[2] = 0; //colour blue
  colGrn[0] = 0; colGrn[1] = 0; colGrn[2] = 0; //colour green
  colRed[0] = 0; colRed[1] = 0; colRed[2] = 0; //colour red
  cmdMode = ' '; // command mode G-gun, R-radar
  cmdType = ' '; // command mode type
  cmdVal = 0; // value associated with a cmdType
  distance = 0; // distance measured by radar
  doWait = 0; // pause in 10ms increments during do commands
  duration = 0;  // measured time delay from echo response
  echoMax = 5000;  // max pulse timeout 5 ms
  echoPin = 4;   // ultrasonic echo pin
  fireDuration = 0; // duration of a fire shot being played in 10ms
  firePnt = -1; // set >= 0 to fire gun sequence
  gunActEn = 0; // gun active/inactive flag
  gunCtrCnt = gunCtrCntDef; // use to slowly centralise gun
  interval = 10000; // main loop interval in microseconds
  keyChar; // any keyboard character
  keyVal = -1; // any keyboard value
  ledCnt = 0; // RGB LED sequence counter
  ledCntSet = 20; // reset value for ledCnt
  ledMode = 0; // RGB LED mode state
  loadFire; // pointer used to load arrays
  loadPnt = 0; // pointer used to load arrays
  nextMicros; // main loop interval in microseconds
  noteDuration = 0; // duration of a note being played in 10ms
  noteDurationVal; // reference duration of a note being played
  notePause; // inter-note pause counter
  numFires = 0;  // number of fires. Must be less than fireSize
  numNotes = 0;  // number of notes in a melody. Must be less than arraySize
  playPnt = -1; // set >= 0 to play melody sequence
  radarCnt = 10; // counter load value for radar trig time-out in 10ms
  radarFlag = 0; // = 1 if sweeping up, -1 if sweeping down
  radarPnt = 90; // current gun sweep value
  radarTrig = 0; // radar trigger/counter
  servoGCtr = 90; // gun centre servo angle limit
  servoGEn = 0; // gun servo enable, 0 = OFF, 1 = ON
  servoGMax = servoGMaxMax; // gun max servo angle limit
  servoGMin = servoGMinMin; // gun min servo angle limit
  sevoGOff = 10; // offset required to achieve centre alignment
  servoGVal = 90; // latest value sent to the gun servo
  servoRCtr = 90; // radar centre servo angle limit
  servoREn = 0; // radar servo enable, 0 = OFF, 1 = ON
  servoRMax = servoRMaxMax; // radar max servo angle limit
  servoRMin = servoRMinMin; // radar min servo angle limit
  servoROff = 2; // offset require to achieve centre alignment
  servoRVal = 90; // latest value sent to the radar servo
  swDel = 20; // delay between siwtch reads
  swCnt = 0; // number of switch presses
  sweepFlag = 0; // = 1 if sweeping up, -1 if sweeping down
  sweepPnt = 90; // current gun sweep value
  swRead = HIGH; // latest input value
  swState = HIGH; //previous switch state
  swTimeout = 0; // inactivity timeout timer
  swVal = 0; // temporary switch value or inactivity timer
  targetA0 = 0; // first detected target angle
  targetA1 = 0; // lastst detected target angle
  targetAM = 0; // derived target angle
  toneFreq0; // note frequency for single note
  toneFreq1; // target note frequency for swept note
  tonePin = tonePinHi; // default tone pin
  toneSwpF = 0; // = 1 if swept tone is being played
  victoryCnt = 0; // >1 count down then play victory tune when = 1
}

// ----------------------------------------------------------------------

void setDP(int zcmdVal) {
  // set the degrees/pulse counter value
  if (zcmdVal > 10) {zcmdVal = 10;}
  if (zcmdVal < 1) {zcmdVal = 1;}
  radarCnt = zcmdVal;
  Serial.print("#DP="); Serial.print(radarCnt); Serial.println(".");
}

// ----------------------------------------------------------------------

void setGunServoOFF() {
  // disable gun servo
  sweepFlag = 0;
}

// ----------------------------------------------------------------------

void setFireLft(int zcmdVal) {
  // set the brightness of the left Fire LED  
  analogWrite(gunLEDLft, zcmdVal);
}

// ----------------------------------------------------------------------

void setFireRht(int zcmdVal) {
  // set the brightness of the left Fire LED  
  analogWrite(gunLEDRht, zcmdVal);
}

// ----------------------------------------------------------------------

void setGunActive(int zcmdVal) {
  // set gun active/inactive mode
  if (zcmdVal > 0) {
    gunActEn = 1;
  } else {
    gunActEn = 0;
  } moveGunServo(servoGCtr);
}

// ----------------------------------------------------------------------

void setGunServoMax(int zcmdVal) {
  // set the maximum gun angle limit
  if (zcmdVal < servoGMinMin) {zcmdVal = servoGMinMin;}
  if (zcmdVal > servoGMaxMax) {zcmdVal = servoGMaxMax;}
  servoGMax = zcmdVal;
  Serial.print("servoGMax=");
  Serial.print(servoGMax); Serial.println(".");
}

// ----------------------------------------------------------------------

void setGunServoMin(int zcmdVal) {
  // set the minimum gun angle limit
  if (zcmdVal > servoGMaxMax) {zcmdVal = servoGMaxMax;}
  if (zcmdVal < servoGMinMin) {zcmdVal = servoGMinMin;}
  servoGMin = zcmdVal;
  Serial.print("servoGMin=");
  Serial.print(servoGMin); Serial.println(".");
}

// ----------------------------------------------------------------------

void setMode(int zMode) {
  /* called to set the mode from mode switch presses
  */
  switch (zMode) {
    case 1:
      // inactive sparkle mode
      ledCntSet = 10; ledCnt = ledCntSet; ledMode = 90;
      break;
    case 2:
      // radar scan only
      setGunActive(0);
      radarFlag = 1; radarPnt = servoRVal;
      radarTrig = radarCnt;
      ledCntSet = 20; ledCnt = ledCntSet; ledMode = 0;
      break;
    case 3:
      // radar scan with active gun
      setGunActive(1);
      radarFlag = 1; radarPnt = servoRVal;
      radarTrig = radarCnt;
      ledCntSet = 20; ledCnt = ledCntSet; ledMode = 3;
      break;
  }
}

// ----------------------------------------------------------------------

void setPixelBlu(uint8_t zVal) {
  // set the pixel 0 Blue colour to zVal
  colBlu[0] = zVal;
  strip.setPixelColor(0,strip.Color(colRed[0], colGrn[0], colBlu[0]));
  strip.show();
}

// ----------------------------------------------------------------------

void setPixelGrn(uint8_t zVal) {
  // set the pixel 0 Green colour to zVal
  colGrn[0] = zVal;
  strip.setPixelColor(0,strip.Color(colRed[0], colGrn[0], colBlu[0]));
  strip.show();
}

// ----------------------------------------------------------------------

void setPixelOFF(){
  // turns OFF all LED values
  setPixiCol(0,0,0,0);
  setPixiCol(1,0,0,0);
  setPixiCol(2,0,0,0);
}

// ----------------------------------------------------------------------

void setPixelRed(uint8_t zVal) {
  // set the pixel 0 Red colour to zVal
  colRed[0] = zVal;
  strip.setPixelColor(0,strip.Color(colRed[0], colGrn[0], colBlu[0]));
  strip.show();
}

// ----------------------------------------------------------------------

void setPixiCol(int zP, uint8_t zRed, uint8_t zGrn, uint8_t zBlu) {
  // set the pixel 0 Red colour to zVal
  colRed[zP] = zRed; colGrn[zP] = zGrn; colBlu[zP] = zBlu;
  strip.setPixelColor(zP,strip.Color(colRed[zP], colGrn[zP], colBlu[zP]));
  strip.show();
}

// ----------------------------------------------------------------------

void setRadServoMax(int zcmdVal) {
  // set the maximum radar angle limit
  if (zcmdVal < servoRMinMin) {zcmdVal = servoRMinMin;}
  if (zcmdVal > servoRMaxMax) {zcmdVal = servoRMaxMax;}
  servoRMax = zcmdVal;
  Serial.print("servoRMax=");
  Serial.print(servoRMax); Serial.println(".");
}

// ----------------------------------------------------------------------

void setRadServoMin(int zcmdVal) {
  // set the minimum radar angle limit
  if (zcmdVal > servoRMaxMax) {zcmdVal = servoRMaxMax;}
  if (zcmdVal < servoRMinMin) {zcmdVal = servoRMinMin;}
  servoRMin = zcmdVal;
  Serial.print("servoRMin=");
  Serial.print(servoRMin); Serial.println(".");
}

// ----------------------------------------------------------------------

void setRadServoOFF() {
  // disable radar servo
  radarFlag = 0;
}

// ----------------------------------------------------------------------

void setToneFreq(unsigned int zFreq) {
  // set the tone output to a given frequency, 0 = no tone
  if (zFreq > 0) {
    tone(tonePin,zFreq); toneFreq = zFreq;
//    Serial.print("ToneFreq=");
//    Serial.println(zFreq);
  } else {noTone(tonePin);}
  toneFreq = zFreq;
}

// ----------------------------------------------------------------------

void sweepGun() {
  // called from main loop, sweeps gun between lower/uppoer limits  
  if (sweepFlag > 0) {
    // sweeping up
    sweepPnt++;
    if (sweepPnt > servoGMax) {sweepPnt = servoGMax;}
    if (sweepPnt == servoGMax) {sweepFlag = -1;}
  }
  else {
    // sweeping down
    sweepPnt--;
    if (sweepPnt < servoGMin) {sweepPnt = servoGMin;}
    if (sweepPnt == servoGMin) {sweepFlag = 1;}
  }
  servoGVal = sweepPnt; gunServo.write(180 - sweepPnt + sevoGOff);
}

// ----------------------------------------------------------------------

void sweepRadar() {
  // called from main loop, sweeps radar between lower/uppoer limits  
  if (radarFlag > 0) {
    // sweeping up
    radarPnt++;
    if (radarPnt > servoRMax) {radarPnt = servoRMax;}
    if (radarPnt == servoRMax) {
      radarFlag = -1;
      if (gunActEn == 2) {gunActEn = 3;}
    }
  }
  else {
    // sweeping down
    radarPnt--;
    if (radarPnt < servoRMin) {radarPnt = servoRMin;}
  }
  servoRVal = radarPnt; radServo.write(180 - radarPnt + servoROff);
  if (radarPnt == servoRMin) {
    radarFlag = 1; Serial.println(">-1,0.");}
    if (gunActEn == 2) {gunActEn = 3;}
  else if (radarTrig < 1) {Serial.print(">-1,");Serial.print(servoRVal);
    Serial.println(".");}
}

// ----------------------------------------------------------------------

void toggleGunSweep() {
  // switch radar sweep ON/OFF
  if (sweepFlag == 0) {
    sweepFlag = 1; sweepPnt = servoGVal;  
  }
  else {sweepFlag = 0;}
}

// ----------------------------------------------------------------------

void toggleRadSweep() {
  // switch radar sweep ON/OFF
  if (cmdVal > 0) {
    // received RS1. so switch it ON
    radarFlag = 1; radarPnt = servoRVal;  
  } else {
    // received RS0. so switch it OFF
    radarFlag = 0;}
}

// ----------------------------------------------------------------------

void toggleRadarTrig() {
  // switch ultrasonic radar triggering ON/OFF
  if (cmdVal > 0) {
    // received U1. so switch it ON
    radarTrig = radarCnt; // switch ON utrasonic triggering
  } else {radarTrig = 0;} // switch OFF utrasonic triggering
}

// ----------------------------------------------------------------------

void trigRadar() {
  // performs an ultrasonic measurement  
  // trigger the distance sensor and get a reading
  // The ultrasonic PING is triggered by a HIGH pulse of 2 or more microseconds.
  digitalWrite(trigPin, LOW);
  delayMicroseconds(2);
  digitalWrite(trigPin, HIGH);
  delayMicroseconds(10); // normally 10us
  digitalWrite(trigPin, LOW);
  
  // A high pulse of duration in microseconds from the sending
  // of the ping to the return of its echo from something.
  duration = pulseIn(echoPin, HIGH, echoMax);
  if (duration == 0) {duration = echoMax - 600;}
  distance = duration;
  // print results in format used by Processing
  Serial.print(">"); Serial.print(distance);
  Serial.print(","); Serial.print(servoRVal);
  Serial.println(".");
  if (gunActEn > 0) {
    // gun is active so move to close targets
    if ((distance <= AlertMin) && (gunActEn < 3)) {
      // target is within danger zone so align gun
      switch (gunActEn) {
        case 1: gunActEn = 2; targetA0 = servoRVal; targetA1 = servoRVal; break;
        case 2: targetA1 = servoRVal; break;
      }
    } else if (gunActEn > 1) {
      // end of target detection
      targetAM = min(targetA0, targetA1);
      targetA1 = max(targetA0, targetA1);
      targetAM = targetAM + ((targetA1 - targetAM)/2);
      moveGunServo(targetAM); loadTuneFire(); victoryCnt = 360;
      Serial.print("Fire="); Serial.println(targetAM);
      targetA0 = 0; targetA1= 0; targetAM = 0; gunActEn = 1;
    }
    else if (distance <= AlertMax) {
      // target is within warning zone so give warning!
      if (playPnt == -1) {loadTuneWarning();}
    }
  }
  if (ledCnt > 0) {
    ledCntSet = (((distance - 2000) * 18)/2400) + 5;
    if (ledCntSet < 3) {ledCntSet = 3;}
//    Serial.println(ledCntSet);
  }
}





